/*
	del_stat.c - statistical udf for InterBase or FireBird
	Idea:
	CreateHandle
	AddData
	Compute
	FreeHandle

	Copyright (C) 2002  Slavomir Skopalik

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

e-mail:skopalik@hlubocky.del.cz
http://hlubocky.del.cz/udf

*/
/*

TODO:
2. StatMeanTimeStamp - replace memalloc by FB internal function
3. StatGetItem

*/
#define DEL_STAT_C
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include <string.h> /* replaced mem.h because can't compile under Linux */
#include "ibase.h"
#include "del_stat.h"

USERES("del_stat.res");
//---------------------------------------------------------------------------
struct TStat {
  int count, datacount;   /* data items, allocated items */
  int sorted;
  int meancomputed;
  int m2computed;
  double mean,min,max;
  double m2;             // Sum of x^2
  double* data;
};

long double powsum(struct TStat *p,double exp){
  int i;
  long double d;
  d=0;
  for(i=0;i<p->count;i++) d+=pow(p->data[i],exp);
  return d;
}

EXPORT(int, StatInit()){
  int count=32;
  return StatInitEx(&count);
}

EXPORT(int, StatInitEx(pint cnt)){
  struct TStat *p;
  p = (struct TStat*) malloc(sizeof(struct TStat));
  if(!p) return 0;                            // Can't allocate memory
  memset(p,0,sizeof(struct TStat));
  p->datacount=*cnt;
  p->data =(double*) malloc(p->datacount * sizeof(double));
  if(!p->data) { free(p); return 0; }         // Can't allocate memory
  return (int)p;
}

EXPORT(int, StatFree(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  free(p->data);
  free(p);
  return 0;
}

EXPORT(int, StatAdd(pint handle,pdouble value)){  // Return index of new item
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(p->count>=p->datacount){
		  p->data=(double*)realloc(p->data,p->datacount*2*sizeof(double));
		  p->datacount*=2;
  };
  p->data[p->count]=*value;
  p->sorted=0; p->meancomputed=0; p->m2computed=0;
  return p->count++;
}

EXPORT(int, StatAddTimeStamp(pint handle, ISC_TIMESTAMP *tm)){
  double d;
  d=tm->timestamp_date+(double)tm->timestamp_time/(double)ISC_TIME_SECONDS_PRECISION/(3600*24);
  return StatAdd(handle,&d);
}

EXPORT(double, StatGetItem(pint handle, pint index)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(*index>=p->count) return 0;
  return p->data[*index];
}

EXPORT(ISC_TIMESTAMP*, StatGetItemTimeStamp(pint handle, pint index)){
  ISC_TIMESTAMP *tm;
  double d;
  tm=(ISC_TIMESTAMP*)malloc(sizeof(ISC_TIMESTAMP));
  d=StatGetItem(handle,index);
  tm->timestamp_date=floor(d);
  tm->timestamp_time=(d-tm->timestamp_date)*ISC_TIME_SECONDS_PRECISION*3600*24;
  return tm;
}

EXPORT(int, StatCount(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  return p->count;
}

EXPORT(int, StatClear(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  p->count=0;
  p->sorted=0; p->meancomputed=0; p->m2computed=0;
  return *handle;
}

EXPORT(int, StatDuplicate(pint handle)){
  struct TStat *p,*p2;
  double *pt;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  p2=(struct TStat*)StatInitEx(&p->count);
  pt=p2->data;
  memmove(p2,p,sizeof(sizeof(struct TStat)));
  p2->data=pt;
  memmove(p2->data,p->data,sizeof(double)*p->count);
  return (int)p2;
}
/*
  Compute Root Mean Squadre and returns result in NEW Stat object
  result form is:
  p[0] + x*p[1] + x^2*p[2] ... x^degree*p[degree]
*/
EXPORT(int, StatLS(pint XValues, pint YValues, pint degree)){
  struct TStat *px, *py, *hr;
  long double **m,d,*t;
  int i,j,r;
  px=(struct TStat*)*XValues; py=(struct TStat*)*YValues;
  if(!px||!py||*degree<0) return 0;
  if(px->count!=py->count) return 0;
  m=(long double**)malloc(sizeof(*m)*(*degree+1));
  for(i=0;i<=*degree;i++) m[i]=(long double*)malloc(sizeof(**m)*(*degree+2));
  for(i=0;i<=*degree;i++) m[0][i]=powsum(px,i);
  for(i=1;i<=*degree;i++){
	for(j=0;j<*degree;j++) m[i][j]=m[i-1][j+1];
	m[i][*degree]=powsum(px,*degree+i);
  };
  for(i=0;i<=*degree;i++){
	d=0;
	for(j=0;j<py->count;j++) d+=py->data[j]*pow(px->data[j],i);
	m[i][*degree+1]=d;
  }
  hr=NULL;
  for(r=0;r<*degree;r++){             // Gaus - Jordan Eliminating method
	d=fabs(m[r][r]); j=r;             // PIVOT search
	for(i=r;i<=*degree;i++)
	  if(fabs(m[i][r])>d){ d=fabs(m[i][r]); j=i; }
	t=m[j]; m[j]=m[r]; m[r]=t;        // Move pivot
	if(d<DBL_MIN) goto err;           // Pivot is too small (probably linear combination of eqution of system)
	for(i=r+1;i<=*degree;i++){
	  d=-m[i][r]/d; //PIVOT
	  for(j=r+1;j<=*degree+1;j++) m[i][j]+=d*m[r][j];
	}
  }                        // Upper triangle matrix computed
  for(r=*degree;r>0;r--){
	if(fabs(m[r][r])<DBL_MIN) goto err;
	for(i=r-1;i>=0;i--){
	  d=-m[i][r]/m[r][r];
	  for(j=r+1;j<=*degree+1;j++) m[i][j]+=d*m[r][j];
	}
  }                      // Diagonal matrix computed
  i=*degree+1;
  hr=(struct TStat*)StatInitEx(&i);
  for(r=0;r<=*degree;r++) hr->data[r]=m[r][*degree+1]/m[r][r];  // Normalize right side
  hr->count=i;
  err:
  for(i=0;i<=*degree;i++) free(m[i]);
  free(m);
  return (int)hr;
}
/*
  Compute value of polynom by Horner optimal schema
  ((a[n]*X + a[n-1])*X + a[n-2])*X + a[n-3] ...
*/
EXPORT(int, StatPoly(pint handle,pdouble X)){
  struct TStat *p;
  int i;
  double d;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  d=0;
  for(i=p->count-1;i>=0;i--) d=d**X+p->data[i];
  return d;
}

int cdecl sort_double( const void *a, const void *b)
{
  if(*(double*)a>*(double*)b) return 1;
  if(*(double*)a<*(double*)b) return -1;
  return 0;
}

EXPORT(int,StatSort(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(p->sorted) return p->count;    
  qsort((void*)p->data,p->count,sizeof(double),sort_double);
  p->sorted=1;
  return p->count;
}

EXPORT(ISC_TIMESTAMP* ,StatMedianTimeStamp(pint handle)){
  ISC_TIMESTAMP *tm;
  double d;
  tm=(ISC_TIMESTAMP*)malloc(sizeof(ISC_TIMESTAMP));
  d=StatMedian(handle);
  tm->timestamp_date=floor(d);
  tm->timestamp_time=(d-tm->timestamp_date)*ISC_TIME_SECONDS_PRECISION*3600*24;
  return tm;
}

EXPORT(double,StatMedian(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(!p->count) return 0;
  if(!p->sorted) StatSort(handle);
  if(p->count&1) return p->data[p->count>>1];
   else return (p->data[(p->count>>1)-1]+p->data[p->count>>1])/2;
}

EXPORT(double,StatMean(pint handle)){
  struct TStat *p;
  int i;
  double d,m,l,h;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(!p->meancomputed) {
	if(!p->count) return 0;
	m=0; l=DBL_MAX; h=-DBL_MAX;
	for(i=0;i<p->count;i++){
	  d=p->data[i];
	  m+=d;
	  if(d<l)l=d;
	  if(d>h)h=d;
	};
	p->mean=m/p->count; p->meancomputed=1; p->min=l; p->max=h;
  };
  return p->mean;
}

EXPORT(ISC_TIMESTAMP* ,StatMeanTimeStamp(pint handle)){
  ISC_TIMESTAMP *tm;
  double d;
  tm=(ISC_TIMESTAMP*)malloc(sizeof(ISC_TIMESTAMP));
  d=StatMean(handle);
  tm->timestamp_date=floor(d);
  tm->timestamp_time=(d-tm->timestamp_date)*ISC_TIME_SECONDS_PRECISION*3600*24;
  return tm;
}

EXPORT(double,StatSum(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  return p->count*StatMean(handle);
}

EXPORT(double,StatMin(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  StatMean(handle);  /* Compute Mean, Min and Max */
  return p->min;
}

EXPORT(double,StatMax(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  StatMean(handle);  /* Compute Mean, Min and Max */
  return p->max;
}

EXPORT(double,StatRange(pint handle)){
  return StatMax(handle)-StatMin(handle);
}

EXPORT(double,StatVariance(pint handle)){
  struct TStat *p;
  int i;
  double m;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(!p->count) return 0;
  if(!p->m2computed){
	m=0;
	for(i=0;i<p->count;i++) m+=p->data[i]*p->data[i];
	p->m2=m;                // seccond common moment
	p->m2computed=1;
  }
  m=StatMean(handle);         // Compute mean if not computed
  return p->m2/p->count-m*m;  // Sum(x^2)/n - (Sum(x)/n)^2
}

EXPORT(double,StatStdDev(pint handle)){
  return sqrt(StatVariance(handle));
}

EXPORT(double,StatVarianceSel(pint handle)){
  struct TStat *p;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(p->count<2) return 0;
  return StatVariance(handle)*p->count/(p->count-1);  // n/(n-1) * Variance
}

EXPORT(double,StatStdDevSel(pint handle)){
  return sqrt(StatVarianceSel(handle));
}

EXPORT(double,StatCentralMoment(pint handle,pint moment)){
  struct TStat *p;
  double d,m;
  int i;
  if(*moment==0) return 1;
  if(*moment==1) return 0;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(*moment==2) return StatVariance(handle);
  m=StatMean(handle);
  d=0;
  for(i=0;i<p->count;i++) d+=pow(p->data[i]-m,*moment);
  return d/p->count;
}

EXPORT(double,StatCommonMoment(pint handle,pint moment)){
  struct TStat *p;
  if(*moment==0) return 1;
  if(*moment==1) return StatMean(handle);
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(*moment==2){StatVariance(handle); return p->m2/p->count;}
  return powsum(p,*moment)/p->count;
}

EXPORT(double,StatSkew(pint handle)){
  struct TStat *p;
  int n3=3;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  return StatCentralMoment(handle,&n3)/pow(StatVariance(handle),3.0/2.0);
}

EXPORT(double,StatSkewSel(pint handle)){     // result = n^2/( (n-1)*(n-2)) * CentralMoment3/VaranceSel^(3/2)
  struct TStat *p;
  int n3=3;
  double n;
  p=(struct TStat*)*handle;
  if(!p) return 0;
  if(p->count<3) return 0;
  n=p->count;
  return n*n*StatCentralMoment(handle,&n3)/pow(StatVarianceSel(handle),3.0/2.0)/(n-1)/(n-2);

}


